/***************************************************************************\
  "k_string.h" - Kevin's String Manipulation Functions
  (Modified for Juliet)
  Revision 66 + 4(Juliet); 3 June 2000 - 9 Feb 2004
  Available at: "http://kevin.alteu.com/"
  E-mail:       "kevin@albrecht.net"

  Copyright (C) 2000-2004, Kevin P. Albrecht

  Permission is hereby granted by Kevin P. Albrecht (the "Author"), free of
  charge, to any person obtaining a copy of this software and associated
  documentation files (the "Software"), to deal in the Software without
  restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the
  Software, and to permit persons to whom the Software is furnished to do
  so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included
  in all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
  THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
  AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

  Except as contained in this notice, the name of the Author shall not be
  used in advertising or otherwise to promote the sale, use or other
  dealings in this Software without prior written authorization from the
  Author.
\***************************************************************************/

#ifndef K_STRING_H
#define K_STRING_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

// External Usage //
#define K_NUMERICS "0123456789-."
#define K_DIGITS "0123456789"
#define K_LETTERS "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
#define K_SPECIAL1 "~`!@#$%^&*()_-+={[}]|\\:;\"'<,>.?/"
#define K_SPECIAL2 "*()-+=:;\"<,>./&"
#define K_LOWERCASE "abcdefghijklmnopqrstuvwxyz"
#define K_UPPERCASE "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
#define K_WHITESPACE "\t\0\n "

// Internal Usage Only //
#define K_STRING_K_SWAP_STRING_SIZE 256
#define K_STRING_K_FLOAT_STR_SIZE 256
#define K_STRING_K_INT_STR_SIZE 256
#define K_STRING_K_POSS_TEMP_SIZE 256
#define K_ERROR_TERMINATE 0
#define K_ERROR_RECOVER 1

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

// Boolean definition //
typedef enum
{
  false, true
} bool;

char *k_app (char *one, char *two);
void k_cat (char *one, char *two);
void k_catd (char *one, char *two);
char k_char (char *str, int pos);
char *k_clone (char *as);
void k_copy (char *one, char *two);
void k_copyd (char *one, char *two);
void k_del (char *str);
void k_error (short int final_status, char *s);
char *k_float (float f, int places);
bool k_inSet (char ch, char *str);
char *k_int (int n);
char *k_new (int size);
int k_posc (char *str, char ch, int after);
int k_poss (char *str, char *sub, int after);
bool k_same (char *one, char *two);
char *k_sub (char *str, int one, int two);
char *k_subd (char *str, int one, int two);
char *k_swap (char *str, char *one, char *two);
char *k_trim (char *str);
char *k_trimd (char *str);

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

//// k_app
//// Outputs the concatenation of 'one' and 'two'
//// If the inputs are NULL, they are given an empty string
char *k_app (char *one, char *two)
{
  char *temp;
  if (one == NULL) strcpy (one, "");
  if (two == NULL) strcpy (two, "");

  temp = k_new (strlen(one) + strlen(two));
  k_copy (temp, one);
  k_cat (temp, two);
  return temp;
}

//// k_cat
//// Concatenates 'two' to the end of 'one'
void k_cat (char *one, char *two)
{
  strcat (one, two);
}

//// k_catd
//// Concatenates 'two' to the end of 'one', then deletes 'two'
void k_catd (char *one, char *two)
{
  strcat (one, two);
  k_del (two);
}

//// k_char
//// Returns the char in the specified position in the specified string
char k_char (char *str, int pos)
{
  if ((pos < 0) || (pos >= (signed) strlen (str)))
  {
    // out of bounds //
    return 0;
  }
  return str[pos];
}

//// k_clone
char *k_clone (char *as)
{
  char *temp;

  if (as != NULL)
  {
    if (as[0] == 0) return NULL; // BUGFIX

    temp = k_new (strlen(as) + 1);
    k_copy (temp, as);
    return temp;
  }
  else
  {
    // BUGFIX (13 May 2001): previously, attempts
    // ... to clone NULL ended in a crash.
    return NULL;
  }
}

//// k_copy
//// Copies 'two' into 'one'
void k_copy (char *one, char *two)
{
  if (two == NULL) strcpy (one, "");
  else             strcpy (one, two);
}

//// k_copyd
//// Copies 'two' into 'one', then deletes 'two'
void k_copyd (char *one, char *two)
{
  if (two == NULL) strcpy (one, "");
  else             strcpy (one, two);
  k_del (two);
}

//// k_del
//// Deletes the string if it is not null or of zero len
void k_del (char *str)
{
  if (str == NULL) return; //Rev 52.6//

  if (strlen (str) > 0)
  {
    JFREE(str);
  }

  // THIS MIGHT CAUSE AN ERROR //
  //JFREE(str);
}

//// k_error
//// Signal a VERY BAD error
void k_error (short int final_status, char *s)
{
  if (final_status == K_ERROR_TERMINATE)
  {
    printf ("K_ERROR: TERMINAL ERROR IN: %s\n", s);
    exit (0);
  }
  else if (final_status == K_ERROR_RECOVER)
  {
    printf ("K_ERROR: RECOVERABLE ERROR IN: %s\n", s);
  }
  else
  {
    printf ("K_ERROR: ERROR #%d IN: %s\n", final_status, s);
  }
}

//// k_float
//// Returns the string equivalent of the float
char *k_float (float f, int places)
{
  int dec, sign, i;
  char *s;
  char *buf;

  if (f == 0) return k_clone("0");

  #ifdef ENV_GCC
    s = (char *) fcvt ((double) f, places, &dec, &sign);
  #endif
  #ifdef ENV_LCC
    s = (char *) _ecvt ((double) f, places, &dec, &sign);
  #endif
  #ifdef ENV_MSVC
    s = (char *) _fcvt ((double) f, places, &dec, &sign);
  #endif

  buf = k_new (K_STRING_K_FLOAT_STR_SIZE);
  if (sign != 0) k_copy (buf, "-");
  else           k_copy (buf, "");

  if (dec <= 0)
  {
    k_cat (buf, ".");
    while (dec < 0)
    {
      k_cat (buf, "0");
      dec++;
    }
    k_cat (buf, s);
  }
  else
  {

    k_catd (buf, k_sub (s, 0, dec));
    k_cat (buf, ".");
    k_catd (buf, k_sub (s, dec, strlen (s)));
  }

  for (i = strlen (buf) - 1; i >= 0; i--)
  {
    switch (buf[i])
    {
      case '0':
        buf[i] = '\0';
        break;
      case '.':
        buf[i] = '\0';
        goto OUT;
      default:
        goto OUT;
    }
  }
  OUT:

  return buf;
}

//// k_inSet
//// Returns true if 'ch' is in 'str'
bool k_inSet (char ch, char *str)
{
  if (k_posc (str, ch, 0) < 0)
    return false;
  else
    return true;
}

//// k_int
//// Returns the string equivalent of the integer
char *k_int (int n)
{
  char *s = k_new (K_STRING_K_INT_STR_SIZE);
  k_copy (s, "");
  itoa (n, s, 10);
  return s;
}

//// k_new
//// Allocate memory for a string; if size=0, call k_error
char *k_new (int size)
{
  void *temp;

  //NOTE - Rev 40 Dec 11, 2001
  //     - 1 is added to the size to leave space for NULL char
  if (size <= 0) k_error (K_ERROR_TERMINATE, "k_new");
  temp = (char *) JALLOC( sizeof(char) * size + 1);

  return temp;
}

//// k_posc
//// Returns the position of the first occurance of the char after 'after';
//// returns -1 if not found
int k_posc (char *str, char ch, int after)
{
  int i;

  if ((after >= (signed) strlen (str)) || (after < 0))
  {
    // out of bounds //
    return -1;
  }

  for (i = after; i < (signed) strlen (str); i++)
  {
    if (str[i] == ch)
    {
      // found //
      return i;
    }
  }

  // not found //
  return -1;
}

//// k_poss
//// Returns the position of the first occurance of the string after 'after';
//// returns -1 if not found
int k_poss (char *str, char *sub, int after)
{
  int str_len, sub_len, temp_len, str_pos, sub_pos;
  char temp[K_STRING_K_POSS_TEMP_SIZE];

  if ((after >= (signed) strlen (str)) || (after < 0) || ((signed) strlen (sub) > (signed) strlen (str)))
  {
    // out of bounds //
    return -1;
  }

  // variables //
  str_len = strlen (str);
  sub_len = strlen (sub);
  temp_len = 0;

  // determine if substring is present:
  for (str_pos = after; str_pos <= str_len; str_pos++)
  {
    if (str[str_pos] == sub[0])
    {
      for (sub_pos = 0; sub_pos < sub_len; sub_pos++)
      {
        temp[sub_pos] = str[str_pos + sub_pos];
        temp_len = sub_pos;
      }
      temp[temp_len + 1] = '\0';
      if (k_same (temp, sub))
      {
        return str_pos;
      }
    }
  }

  return -1;
}

//// k_same
//// Returns true if 'one' and 'two' are the same
bool k_same (char *one, char *two)
{
  if ((one[0] == 0) || (two[0] == 0)) k_error (K_ERROR_TERMINATE, "k_same");
  return (strcmp (one, two) == 0);
}

//// k_sub
//// Returns the substring between two points; '0' if out of bounds
char *k_sub (char *str, int one, int two)
{
  int str_len, str_pos, sub_pos;
  char *sub;

  str_len = strlen (str);

  if ((one < 0) || (one > str_len) || (two < one))
  {
    // out of bounds //
    return 0;
  }
  if (one == two)
  {
    // zero length //
    return "";
  }

  sub = k_new (two - one + 1);
  strcpy (sub, "");

  // create substring //
  for (str_pos = one, sub_pos = 0; str_pos < two; str_pos++, sub_pos++)
  {
    sub[sub_pos] = str[str_pos];
  }
  sub[two - one] = '\0'; // end with '\0'

  return sub;
}

//// k_sub
//// Returns the substring between two points; '0' if out of bounds;
////  deletes the argument
char *k_subd (char *str, int one, int two)
{
  int str_len, str_pos, sub_pos;
  char *sub;

  str_len = strlen (str);
  sub = k_new (two - one + 1);
  strcpy (sub, "");

  if ((one < 0) || (one > str_len) || (two < one))
  {
    // out of bounds //
    k_del (sub);
    k_del (str);
    return 0;
  }
  if (one == two)
  {
    // zero length //
    k_del (sub);
    k_del (str);
    return "";
  }

  // create substring //
  for (str_pos = one, sub_pos = 0; str_pos < two; str_pos++, sub_pos++)
  {
    sub[sub_pos] = str[str_pos];
  }
  sub[two - one] = '\0'; // end with '\0'

  k_del (str);
  return sub;
}


//// k_swap
//// Swaps all the characters in a string that occur in the first array with matching
////  characters in the second array; returns the string with swapped characters;
////  '0' if errors
char *k_swap (char *str, char *one, char *two)
{
  int str_len, array_len, str_pos, array_pos;
  char *swapped;

  str_len = strlen (str);
  array_len = strlen (one);
  swapped = k_new (K_STRING_K_SWAP_STRING_SIZE); /*str_len might be zero*/

  strcpy (swapped, str);

  if (strlen (one) != strlen (two))
  {
    // arrays do not match //
    k_del (swapped);
    return 0;
  }

  for (str_pos = 0; str_pos < str_len; str_pos++)
  {
    for (array_pos = 0; array_pos < array_len ; array_pos++)
    {
      if (swapped[str_pos] == one[array_pos])
      {
        swapped[str_pos] = two[array_pos];
      }
    }
  }

  return swapped;
}

//// k_trim
//// Returns the string trimmed of white space
char *k_trim (char *str)
{
  int end, start;
  char *temp;

  end = strlen (str);
  start = 0;

  while ((start < end) && (str[start] <= ' '))
  {
    start++;
  }

  while ((start < end) && (str[end - 1] <= ' '))
  {
    end--;
  }

  if ((start > 0) || (end < (signed) strlen (str)))
  {
    return k_sub (str, start, end);
  }
  else
  {
    temp = k_new (strlen (str));
    strcpy (temp, str);
    return temp;
  }
}

//// k_trim
//// Returns the string trimmed of white space; deletes the arg
char *k_trimd (char *str)
{
  int end, start;
  char *temp;

  end = strlen (str);
  start = 0;

  while ((start < end) && (str[start] <= ' '))
  {
    start++;
  }

  while ((start < end) && (str[end - 1] <= ' '))
  {
    end--;
  }

  if ((start > 0) || (end < (signed) strlen (str)))
  {
    temp = k_subd (str, start, end);
    return temp;
  }
  else
  {
    return str;
  }
}

#endif
